/*
 * SonarQube
 * Copyright (C) 2009-2016 SonarSource SA
 * mailto:contact AT sonarsource DOT com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
package org.sonar.api.platform;

import java.io.File;
import java.util.Date;
import javax.annotation.CheckForNull;
import org.sonar.api.batch.ScannerSide;
import org.sonar.api.ce.ComputeEngineSide;
import org.sonar.api.server.ServerSide;

/**
 * Runtime information about server
 *
 * @since 2.2
 */
@ScannerSide
@ServerSide
@ComputeEngineSide
public abstract class Server {

  /**
   * Name is misleading, this is an UUID generated
   * at each startup, so it changes if server is restarted.
   * @return a non-null UUID. Format can change over versions.
   */
  public abstract String getId();

  /**
   * UUID generated on demand by system administrators. It is
   * {@code null} by default on fresh installations. When defined,
   * value does not change when server is restarted.
   * @since 2.10
   */
  @CheckForNull
  public abstract String getPermanentServerId();

  /**
   * Non-null version of SonarQube at runtime
   */
  public abstract String getVersion();

  /**
   * Date when server started
   */
  public abstract Date getStartedAt();

  /**
   * @deprecated in 6.0. Replaced by {@link ServerFileSystem#getHomeDir()}
   * @return an existing directory in server and CE environments, {@code null} in scanner.
   */
  @Deprecated
  public abstract File getRootDir();

  /**
   * @deprecated always {@code null} since version 6.0. No alternatives, as plugins do not have to touch this directory.
   */
  @Deprecated
  @CheckForNull
  public abstract File getDeployDir();

  /**
   * Context path of web server. Value is blank {@code ""} by default. When defined by
   * the property {@code sonar.web.context} of conf/sonar.properties, then value starts but does
   * not end with slash {@code '/'}, for instance {@code "/sonarqube"}.
   *
   * @return non-null but possibly blank path
   */
  public abstract String getContextPath();

  /**
   * Return the public root url, for instance : https://nemo.sonarqube.org.
   * Default value is {@link org.sonar.api.CoreProperties#SERVER_BASE_URL_DEFAULT_VALUE}
   *
   * @since 5.4
   */
  public abstract String getPublicRootUrl();

  /**
   * The dev mode is enabled when the property {@code sonar.web.dev} is {@code true}.
   *
   * @since 5.4
   */
  public abstract boolean isDev();

  /**
   * Return whether or not the {#getPublicRootUrl} is started with https.
   *
   * @since 5.4
   * @deprecated since 5.6, use instead {@link javax.servlet.http.HttpServletRequest#getHeader(String)} and check that X-Forwarded-Proto header is set to "https".
   */
  @Deprecated
  public abstract boolean isSecured();

  /**
   * @return the server URL
   * @since since 2.4 on batch side only, since 5.6 on both batch side and server side (WebServer and Compute Engine)
   */
  public abstract String getURL();
}
